/*
 * Dynamic loading and ffi
 */

#include "s.h"
#include <dlfcn.h>
#include <avcall.h>

/*-- FFI type marker
 *
 * Note: keywords passed for types: don't need to be gc protected
 * because all keywords are referenced in the keyword hash. They never
 * move or will be removed.
 */

static SOBJ
  EXT_KEYW_ANY,
  EXT_KEYW_VOID,
  EXT_KEYW_CHAR,
  EXT_KEYW_SHORT,
  EXT_KEYW_USHORT,
  EXT_KEYW_INT,
  EXT_KEYW_UINT,
  EXT_KEYW_LONG,
  EXT_KEYW_ULONG,
  EXT_KEYW_FLOAT,
  EXT_KEYW_DOUBLE,
  EXT_KEYW_STATIC_PTR,
  EXT_KEYW_ITEM,				/* alias for static-pointer */
  EXT_KEYW_DYNAMIC_PTR,
  EXT_KEYW_STRING,
  EXT_KEYW_BOOLEAN;

/* This value is stored in the ExtFunc structure. Conversion between
   keyword and type is performed in the make_ext_func */

enum ExtTypes {
  EXT_T_ERROR = -2,
  EXT_T_ANY = -1,
  EXT_T_VOID,
  EXT_T_CHAR,
  EXT_T_SHORT,
  EXT_T_USHORT,
  EXT_T_INT,
  EXT_T_UINT,
  EXT_T_LONG,
  EXT_T_ULONG,
  EXT_T_FLOAT,
  EXT_T_DOUBLE,
  EXT_T_STATIC_PTR,
  EXT_T_DYNAMIC_PTR,
  EXT_T_STRING,
  EXT_T_BOOLEAN,
  EXT_T_MAX
};

/* table used to convert keyword to type */

struct ExtKeywType {
  SOBJ 	*atom;
  short type;
};

static struct ExtKeywType keyw2type_ref[] = {
  { &EXT_KEYW_ANY,  		EXT_T_ANY},
  { &EXT_KEYW_VOID,  		EXT_T_VOID},
  { &EXT_KEYW_CHAR,  		EXT_T_CHAR},
  { &EXT_KEYW_SHORT, 		EXT_T_SHORT},
  { &EXT_KEYW_USHORT,  		EXT_T_USHORT},
  { &EXT_KEYW_INT,  		EXT_T_INT},
  { &EXT_KEYW_UINT,  		EXT_T_UINT},
  { &EXT_KEYW_LONG,  		EXT_T_LONG},
  { &EXT_KEYW_ULONG,  		EXT_T_ULONG},
  { &EXT_KEYW_FLOAT,  		EXT_T_FLOAT},
  { &EXT_KEYW_DOUBLE,  		EXT_T_DOUBLE},
  { &EXT_KEYW_STATIC_PTR,  	EXT_T_STATIC_PTR},
  { &EXT_KEYW_ITEM,			EXT_T_STATIC_PTR},
  { &EXT_KEYW_DYNAMIC_PTR,  EXT_T_DYNAMIC_PTR},
  { &EXT_KEYW_STRING,  		EXT_T_STRING},
  { &EXT_KEYW_BOOLEAN,  	EXT_T_BOOLEAN},
  { NULL },
};

/*-- convert the keyword to an external type */
static int keyword_to_ext_type(SOBJ keyw)
{
  struct ExtKeywType *k = keyw2type_ref;
  
  while(k->atom) {
	if (*k->atom == SCM_KEYW_NAME(keyw))
	  return(k->type);
	k++;
  }
  return(EXT_T_ERROR);
}



#define DLFLAGS	(RTLD_LAZY|RTLD_GLOBAL)

static SOBJ scm_dl_list;

void *scm_find_extsym(char *path, char *sym_name, int must)
{
  void *handle, *sym;
  SOBJ l, str;
  char *error;

  str = scm_mkstring(path);
  
  if (scm_dl_list == NULL) 	err("dl_list not initialized", NULL);
  if ((l = scm_member(str, scm_dl_list)) != scm_false) {

	handle = SCM_POINTER(SCM_CAR(SCM_CDR(l)));
  } else {
	if ((handle=dlopen(path, DLFLAGS)) == NULL) err(dlerror(), str);
	scm_dl_list = scm_cons(str,
						   scm_cons(scm_mk_static_pointer(handle),
									scm_dl_list));
  }

  sym = dlsym(handle, sym_name);

  if (must && (error = dlerror()) != NULL) {
	err("find-func: cannot find func", scm_mkstring(error));
  }
  return(sym);
}

static void load_and_call(char *path, char *sym_name)
{
  void (*init_func)();
  SOBJ str = scm_mkstring(path);

  if ((init_func = scm_find_extsym("", sym_name, FALSE)) != NULL)
	err("load-and-call: module already loaded (statically)", str);

  if (scm_member(str, scm_dl_list) != scm_false)
	err("load-and-call: module already loaded (dynamically)", str);

  init_func = scm_find_extsym(path, sym_name, TRUE);
  (*init_func)();
}

#define INIT_FUNC_PREFIX		"scm_init_"
#define INIT_FUNC_PREFIX_LEN	strlen(INIT_FUNC_PREFIX)

#define INIT_SYM_NAME_MAX		64


static void load_library(char *path)
{
  char init_sym_name[INIT_SYM_NAME_MAX];
  char *p, *q;
  int len;
  
  p = strrchr(path, '/');
  if (p == NULL)				/* no / => current path */
	p = path;
  else 
	p++;
  
  q = strchr(p, '.');			/* search for possible suffix */
  if (q == NULL)				/* no suffix ? set q to end of string */
	q = p + strlen(p);

  len = INIT_FUNC_PREFIX_LEN + q-p;
  if (len >= (INIT_SYM_NAME_MAX-1))
	err("load-library: init symbol too long", NULL);

  strcpy(init_sym_name, INIT_FUNC_PREFIX);
  strncpy(init_sym_name + INIT_FUNC_PREFIX_LEN, p, q-p);
  init_sym_name[len] = 0;
  load_and_call(path, init_sym_name);
}

SOBJ scm_mk_static_pointer(void *p)
{
  SOBJ new = scm_mkpointer();
  SCM_POINTER(new) = p;
  return(new);
}

SOBJ scm_mk_dynamic_pointer(void *p)
{
  SOBJ new = scm_mkpointer();
  SCM_POINTER(new) = p;
  SCM_POINTER_ATTRIB(new) = SCM_POINTER_ALLOCED;
  return(new);
}

SOBJ scm_load_library(SOBJ x)
{
  if (!SCM_STRINGP(x)) err("load-library: bad string", x);
  load_library(SCM_STR_VALUE(x));
  return(scm_true);
}

/****************************************************************
 * FFI
 ****************************************************************/

/*-- gc function for the extfunc */

void scm_extfunc_mark(SOBJ extfunc)
{
}

void scm_extfunc_sweep(SOBJ extfunc)
{
  if (SCM_EXTFUNC(extfunc)) {
	scm_free(SCM_EXTFUNC(extfunc));
  }
  SCM_EXTFUNC(extfunc) = NULL;
}

SOBJ scm_mkextfunc(SCM_ExtFunc *f)
{
  SOBJ obj = scm_newcell(SOBJ_T_EXTFUNC);
  SCM_EXTFUNC(obj) = f;
  return(obj);
}

void scm_extfunc_print(SOBJ x, PORT *p)
{
  port_puts(p, "<#extfunc>");
}

void scm_extfunc_write(SOBJ x, PORT *p)
{
  char buf[128];
  sprintf(buf, "<#extfunc %p: ret=%d func=%p nargs=%d optargs=%d>",
		  x,
		  SCM_EXTFUNC(x)->return_t,
		  SCM_EXTFUNC(x)->func,
		  SCM_EXTFUNC(x)->argc,
		  SCM_EXTFUNC(x)->vararg);
  port_puts(p, buf);
}


static SCM_ExtFunc *scm_new_extfunc()
{
  return(scm_must_alloc(sizeof(SCM_ExtFunc)));
}

/*-- create an external function object.
 *
 * Syntax: (make-external-func <lib> <ret-type> <name> <arglist> )
 * Example: (make-external-function "" 0 'printf '(12 . 0))
 */
static void errext(char *string, SOBJ x)
{ 
  char buf[128];
  sprintf(buf, "make-external-function: %s", string);
  err(buf, x);
}

static void errcall(char *string, SOBJ x)
{ 
  char buf[128];
  sprintf(buf, "call-external-function: %s", string);
  err(buf, x);
}


SOBJ scm_make_extfunc(SOBJ lib, SOBJ ret, SOBJ name, SOBJ argl)
{
  SCM_ExtFunc *f;
  void *func;
  int iret = 0, itype = 0;

  if (!SCM_STRINGP(lib))	errext("bad library name", lib);
  if (!SCM_STRINGP(name))	errext("bad function name", name);
  if (!SCM_KEYWORDP(ret) ||
	  (iret = keyword_to_ext_type(ret)) == EXT_T_ERROR)
	errext("bad return type", ret);

  func = scm_find_extsym(SCM_STR_VALUE(lib), SCM_STR_VALUE(name), FALSE);
  if (func == NULL) 		errext("function not found", scm_cons(lib, name));
  
  f = scm_new_extfunc();
  f->func = func;
  f->return_t = iret;
  f->argc = 0;

  while(argl) {
	if (!SCM_PAIRP(argl)) {
	  f->vararg = 1;
	  break;
	}
	if (!SCM_KEYWORDP(SCM_CAR(argl)) ||
		(itype = keyword_to_ext_type(SCM_CAR(argl))) == EXT_T_ERROR) 
	  errext("bad argument type", SCM_CAR(argl));
	
	f->arg_t[f->argc++] = itype;
	argl = SCM_CDR(argl);
  }
  return(scm_mkextfunc(f));
}

static void pushargs(av_alist *al, SCM_ExtFunc *f, int nargs, SOBJ *arg)
{
  int i;
  SOBJ obj;

  if (f->vararg) {
	if (nargs < f->argc) 	errcall("not enough args", NULL);
  } else {
	if (nargs != f->argc)	errcall("bad number of args", NULL);
  }

  for (i = 0; i < f->argc; i++) {
	obj = arg[i];
	switch(f->arg_t[i]) {
	case EXT_T_VOID:	errcall("void argument not allowed", NULL);
	case EXT_T_CHAR:
	case EXT_T_SHORT:	case EXT_T_USHORT:
	case EXT_T_INT:		case EXT_T_UINT:
	case EXT_T_LONG:	case EXT_T_ULONG:
	  if (SCM_CHARP(obj))  { av_long(*al, SCM_CHAR(obj));  break;}
	  if (SCM_NUMBERP(obj)){ av_long(*al, scm_number2long(obj)); break;}
	  errcall("bad integer", obj);

	case EXT_T_FLOAT:
	  if (SCM_NUMBERP(obj)){ av_float(*al, scm_number2double(obj)); break;}
	  errcall("bad float", obj);

	case EXT_T_DOUBLE:
	  if (SCM_NUMBERP(obj)){ av_double(*al, scm_number2double(obj)); break;}
	  errcall("bad double", obj);

	case EXT_T_STATIC_PTR:	case EXT_T_DYNAMIC_PTR:	case EXT_T_STRING:
	  if (SCM_POINTERP(obj)){ av_ptr(*al,void*,SCM_POINTER(obj)); break;}
	  if (SCM_STRINGP(obj)) { av_ptr(*al,void*,SCM_STR_VALUE(obj));break;}
	  errcall("bad pointer", obj);
	  
	case EXT_T_BOOLEAN:
	  av_long(*al, obj != scm_false);
	  break;

	default:
	  errcall("bad argument type", SCM_MKINUM(f->arg_t[i]));
	}
  }

  /* for variable arguments, try to guess type */
  while(i < nargs) {
	obj = arg[i];
	switch(SCM_OBJTYPE(obj)) {
	case SOBJ_T_INUM:
	case SOBJ_T_BNUM:	av_long(*al, scm_number2long(obj));	  	break;
	case SOBJ_T_FNUM:	av_double(*al, scm_number2double(obj)); break;
	case SOBJ_T_CHAR:	av_char(*al, SCM_CHAR(obj));			break;
	case SOBJ_T_STRING:	av_ptr(*al, void*, SCM_STR_VALUE(obj));	break;
	case SOBJ_T_SYMBOL:	av_ptr(*al, void*, SCM_SYM_NAME(obj));	break;
	case SOBJ_T_POINTER:av_ptr(*al, void*, SCM_POINTER(obj));	break;
	default:
	  if (obj == NULL) {
		av_ptr(*al, void*, NULL);
	  } else {
		errcall("don't know how to convert", obj);
	  }
	}
	i++;
  }
}


SOBJ scm_extfunc_call(SOBJ proc, int nargs, SOBJ *arg)
{
  av_alist alist;
  SCM_ExtFunc *f = SCM_EXTFUNC(proc);

  switch(f->return_t) {
  case EXT_T_VOID:
	av_start_void(alist, f->func);
	pushargs(&alist, f, nargs, arg);
	av_call(alist);
	return(scm_undefined);

  case EXT_T_INT:
	{
	  int result;
	  av_start_int(alist, f->func, &result);
	  pushargs(&alist, f, nargs, arg);
	  av_call(alist);
	  return(scm_int2num(result));
	}
  case EXT_T_DOUBLE:
	{
	  double result;
	  av_start_int(alist, f->func, &result);
	  pushargs(&alist, f, nargs, arg);
	  av_call(alist);
	  return(scm_mkfnum(result));
	}
  case EXT_T_STATIC_PTR:	case EXT_T_DYNAMIC_PTR:
	{
	  void *p;

	  av_start_ptr(alist, f->func, void *, &p);
	  pushargs(&alist, f, nargs, arg);
	  av_call(alist);
	  return( f->return_t == EXT_T_STATIC_PTR ?
			  scm_mk_static_pointer(p) :
			  scm_mk_dynamic_pointer(p) );
	}
	
  default:
	errcall("return type not supported", SCM_MKINUM(f->return_t));
  }
  return(NULL);
}
					  

/*
 * (external-exist? "" "test")
 */

SOBJ scm_external_existsp(SOBJ lib, SOBJ entry)
{
  if (!SCM_STRINGP(entry)) 	err("external-exists?: bad string", entry);
  if (!SCM_STRINGP(lib)) 	err("external-exists?: bad string", lib);

  return(SCM_MKBOOL(scm_find_extsym(SCM_STR_VALUE(lib),
							  SCM_STR_VALUE(entry), FALSE)));
}

/*
 * (call-external "" :int "printf" (:str . :any) ("Hello world %d %d\n" 10 20))
 */
static void push_arg(av_alist *al, int type, SOBJ arg)
{
  switch(type) {
  case EXT_T_VOID:	err("external-call: cannot push void arg", arg);

  case EXT_T_CHAR:
  case EXT_T_SHORT:		case EXT_T_USHORT:
  case EXT_T_INT:		case EXT_T_UINT:
  case EXT_T_LONG:		case EXT_T_ULONG:
	if (SCM_CHARP(arg)) 	{ av_long(*al, SCM_CHAR(arg)); break; }
	if (SCM_NUMBERP(arg)) 	{ av_long(*al, scm_number2long(arg)); break; }
	err("external-call: bad integer", arg);
	  
  case EXT_T_FLOAT:
	if (SCM_CHARP(arg)) 	{ av_float(*al, SCM_CHAR(arg)); break; }
	if (SCM_NUMBERP(arg)) 	{ av_float(*al, scm_number2double(arg)); break; }
	err("external-call: bad float", arg);

  case EXT_T_DOUBLE:
	if (SCM_CHARP(arg)) 	{ av_double(*al,SCM_CHAR(arg)); break; }
	if (SCM_NUMBERP(arg)) 	{ av_double(*al,scm_number2double(arg)); break; }
	err("external-call: bad double", arg);

  case EXT_T_STATIC_PTR:	case EXT_T_DYNAMIC_PTR:	case EXT_T_STRING:
	if (SCM_POINTERP(arg)) 	{ av_ptr(*al, void*, SCM_POINTER(arg)); break;}
	if (SCM_STRINGP(arg))  	{ av_ptr(*al, void*, SCM_STR_VALUE(arg));break;}
	err("external-call: bad pointer", arg);

  case EXT_T_BOOLEAN:
	if (SCM_BOOLEANP(arg)) 	{ av_long(*al, (arg != scm_false)); break; }
	if (SCM_NUMBERP(arg)) 	{ av_long(*al, scm_number2long(arg)); break; }
	err("external-call: bad boolean", arg);
  default:
	err("external-call: bad argument type", arg);
  }
}

static void push_list(av_alist *al, SOBJ list)
{
  SOBJ arg;
  while(list) {
	arg = SCM_CAR(list);
	switch(SCM_OBJTYPE(arg)) {
	case SOBJ_T_INUM:
	case SOBJ_T_BNUM:		av_long(*al, scm_number2long(arg));	  		break;
	case SOBJ_T_FNUM:		av_double(*al, scm_number2double(arg));  	break;
	case SOBJ_T_CHAR:		av_char(*al, SCM_CHAR(arg));			 	break;
	case SOBJ_T_STRING:		av_ptr(*al, void*, SCM_STR_VALUE(arg));		break;
	case SOBJ_T_SYMBOL:		av_ptr(*al, void*, SCM_SYM_NAME(arg));		break;
	case SOBJ_T_POINTER: 	av_ptr(*al, void*, SCM_POINTER(arg));		break;
	default:
	  err("push_list: don't know how to convert", arg);
	}
	list = SCM_CDR(list);
  }
}

static void push_args(av_alist *al, SOBJ argtype, SOBJ argval)
{
  /* prepare arguments */
  while(argtype) {
	if (!SCM_PAIRP(argtype)) {	/* have a rest */
	  push_list(al, argval);
	  break;
	}
	if (argval == NULL)	err("call-external: not enough arguments", argtype);

	if (!SCM_INUMP(SCM_CAR(argtype)))
	  err("call-external: argtype is not a number", SCM_CAR(argtype));

	push_arg(al, SCM_INUM(SCM_CAR(argtype)), SCM_CAR(argval));
	argtype = SCM_CDR(argtype);
	argval  = SCM_CDR(argval);
  }
  
}

SOBJ scm_external_call(SOBJ lib, SOBJ ret, SOBJ entry, 
					   SOBJ argtype, SOBJ argval)
{
  av_alist alist;
  void *func;

  /*-- type checking */
  if (!SCM_STRINGP(lib))	err("external-call: bad library name", lib);
  if (!SCM_STRINGP(entry))	err("external-call: bad entry name", entry);
  if (!SCM_INUMP(ret))		err("external-call: bad return type", ret);

  func = scm_find_extsym(SCM_STR_VALUE(lib), SCM_STR_VALUE(entry), FALSE);
  if (func == NULL)
	err("external-call: cannot find func", scm_cons(lib, entry));

  switch(SCM_INUM(ret)) {
  case EXT_T_VOID:
	av_start_void(alist, func);
	push_args(&alist, argtype, argval);
	av_call(alist);
	return(scm_undefined);

  case EXT_T_INT:
	{
	  int result;
	  av_start_int(alist, func, &result);
	  push_args(&alist, argtype, argval);
	  av_call(alist);
	  return(scm_int2num(result));
	}

  case EXT_T_DOUBLE:
	{
	  double result;
	  av_start_double(alist, func, &result);
	  push_args(&alist, argtype, argval);
	  av_call(alist);
	  return(scm_mkfnum(result));
	}

  default:
	err("external-call: unsupported return type", ret);
  }
  return(scm_undefined);
}

/*-- ExtVar related functions */

void scm_extvar_mark(SOBJ extvar)
{
}

void scm_extvar_sweep(SOBJ extvar)
{
  if (SCM_EXTVAR(extvar)) {
	scm_free(SCM_EXTVAR(extvar));
  }
  SCM_EXTVAR(extvar) = NULL;
}


SOBJ scm_mkextvar(SCM_ExtVar *v)
{
  SOBJ obj = scm_newcell(SOBJ_T_EXTVAR);
  SCM_EXTVAR(obj) = v;
  return(obj);
}

void scm_extvar_print(SOBJ x, PORT *p)
{
  port_puts(p, "<#extvar>");
}

void scm_extvar_write(SOBJ x, PORT *p)
{
  char buf[128];
  sprintf(buf, "<#extvar %p: ptr=%p type=%d attrib=%d>",
		  x,
		  SCM_EXTVAR(x)->ptr,
		  SCM_EXTVAR(x)->type,
		  SCM_EXTVAR(x)->attrib);
  port_puts(p, buf);
}


SCM_ExtVar *scm_new_extvar()
{
  return(scm_must_alloc(sizeof(SCM_ExtVar)));
}

SOBJ scm_make_extvar(SOBJ lib, SOBJ type, SOBJ name)
{
  SCM_ExtVar *extvar;
  void *ptr;
  int itype = 0;
  
  if (!SCM_STRINGP(lib)) err("make-external-var: bad library name", lib);
  if (!SCM_STRINGP(name))err("make-external-var: bad variable name", name);
  if (!SCM_KEYWORDP(type) ||
	  (itype = keyword_to_ext_type(type)) == EXT_T_ERROR)
	err("make-external-var: bad type value", type);
  
  ptr = scm_find_extsym(SCM_STR_VALUE(lib), SCM_STR_VALUE(name), TRUE);
  
  extvar = scm_new_extvar();
  extvar->ptr = ptr;
  extvar->type = itype;
  extvar->attrib = 0;
  return(scm_mkextvar(extvar));
}

SOBJ scm_extvar_get(SOBJ var)
{
  long n;
  void *ptr;
  SOBJ obj;

  if (!SCM_EXTVARP(var)) err("extvar-get: bad external var", var);
  ptr = SCM_EXTVAR(var)->ptr;

  switch(SCM_EXTVAR(var)->type) {
  case EXT_T_CHAR:		return(scm_mkchar( *((char *)ptr) ));
  case EXT_T_SHORT:		n = *((short *)ptr);  return(SCM_MKINUM(n));
  case EXT_T_USHORT:	n = *((unsigned short *)ptr);  return(SCM_MKINUM(n));
  case EXT_T_INT:		n = *((int *)ptr);  return(scm_int2num(n));
  case EXT_T_UINT:		n = *((unsigned int *)ptr);  return(scm_int2num(n));
  case EXT_T_LONG:		n = *((long *)ptr);  return(scm_int2num(n));
  case EXT_T_ULONG:		n = *((unsigned long *)ptr);  return(scm_int2num(n));
  case EXT_T_FLOAT:		return(scm_mkfnum(*((float*)ptr)));
  case EXT_T_DOUBLE:	return(scm_mkfnum(*((double*)ptr)));
  case EXT_T_STATIC_PTR:	
  case EXT_T_DYNAMIC_PTR:
	obj = scm_mkpointer();
	SCM_POINTER(obj) = ptr;
	return(obj);
  case EXT_T_STRING:
	return(scm_mkstring(ptr));
  case EXT_T_BOOLEAN:
	return( SCM_MKBOOL( *((int *)ptr) ));
  default:
	err("extvar-get: bad type", var);
  }
  return(scm_undefined);
}

SOBJ scm_extvar_set(SOBJ var, SOBJ value)
{
  void *ptr;

  ptr = SCM_EXTVAR(var)->ptr;
  switch(SCM_EXTVAR(var)->type) {
  case EXT_T_CHAR:
	if (!SCM_CHARP(value))	err("extvar-set: bad char", value);
	*((char *)ptr) = SCM_CHAR(value);
	break;

  case EXT_T_SHORT:
  case EXT_T_USHORT:	*((short *)ptr) = scm_number2long(value);	break;
	
  case EXT_T_INT:		
  case EXT_T_UINT:		*((int *)ptr) = scm_number2long(value);		break;
  case EXT_T_LONG:	
  case EXT_T_ULONG:		*((long *)ptr) = scm_number2long(value);	break;
  case EXT_T_FLOAT:		*((float*)ptr) = scm_number2double(value);	break;
  case EXT_T_DOUBLE:	*((double*)ptr)= scm_number2double(value);	break;
  case EXT_T_BOOLEAN:	*((int*)ptr) = (value != scm_false);		break;

  case EXT_T_STATIC_PTR:	
  case EXT_T_DYNAMIC_PTR:
  case EXT_T_STRING:
  default:
	err("extvar-get: bad type", var);
  }
  return(scm_undefined);
}


/*-- initialize this module */

void scm_init_dyn()
{
  void *handle = dlopen(NULL, DLFLAGS);
  if (handle == NULL) err("dyn: cannot init", scm_mkstring(dlerror()));

  scm_dl_list = SCM_LIST2(scm_mkstring(""), scm_mk_static_pointer(handle));
  scm_gc_protect(&scm_dl_list);
  scm_add_cvar("library-list", &scm_dl_list);

  EXT_KEYW_ANY			= scm_mkatom(":any");
  EXT_KEYW_VOID			= scm_mkatom(":void");
  EXT_KEYW_CHAR			= scm_mkatom(":char");
  EXT_KEYW_SHORT		= scm_mkatom(":short");
  EXT_KEYW_USHORT		= scm_mkatom(":ushort");
  EXT_KEYW_INT			= scm_mkatom(":int");
  EXT_KEYW_UINT			= scm_mkatom(":uint");
  EXT_KEYW_LONG			= scm_mkatom(":long");
  EXT_KEYW_ULONG		= scm_mkatom(":ulong");
  EXT_KEYW_FLOAT		= scm_mkatom(":float");
  EXT_KEYW_DOUBLE		= scm_mkatom(":double");
  EXT_KEYW_STATIC_PTR	= scm_mkatom(":pointer");
  EXT_KEYW_ITEM			= scm_mkatom(":item");
  EXT_KEYW_DYNAMIC_PTR	= scm_mkatom(":dynamic-ptr");
  EXT_KEYW_STRING		= scm_mkatom(":string");
  EXT_KEYW_BOOLEAN 		= scm_mkatom(":boolean");

  scm_add_cprim("make-extfunc",		scm_make_extfunc, 4);
  scm_add_cprim("load-library", 	scm_load_library, 1);
  scm_add_cprim("external-exists?", scm_external_existsp, 2);
  scm_add_cprim("external-call", 	scm_external_call, 5);

  scm_add_cprim("make-extvar",		scm_make_extvar, 3);
  scm_add_cprim("extvar-get",		scm_extvar_get, 1);
  scm_add_cprim("extvar-set",		scm_extvar_set, 2);
}

